/***************************************************************************
                          configuration.h  -  description
                             -------------------
    begin                : Tue Jun 11 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CONFIGURATION_H
#define CONFIGURATION_H

/**
  *@author Morten Hustveit
  */

#include <qobject.h>
#include <qmap.h>
#include <qptrvector.h>
#include <qvaluevector.h>

#include "midi.h"

class QString;
class KConfig;
class KURL;
class SanetrackerApp;
class ConfigDialog;

class Configuration : public QObject
{
  Q_OBJECT

public:

  Configuration(SanetrackerApp* parent);
  ~Configuration();

  void configure(int defaultTab = 0, bool dirty = false);
  void load();
  void save();

  typedef QMap<int, QString> Bank;
  typedef QMap<int, QString> Control;
  typedef QMap<int, QString> NRPNs;
  typedef QMap<int, QString> RPNs;
  typedef QMap<int, QString> Notes;

  typedef QMap<QString, Bank>       BankMap;
  typedef QMap<QString, Control>    ControlMap;
  typedef QMap<QString, NRPNs>      NRPNMap;
  typedef QMap<QString, RPNs>       RPNMap;
  typedef QMap<QString, Notes>      NoteMap;

  class Instrument
  {
  public:

    enum BankSelMethod
    {
      Normal = 1
    };

    Instrument()
      : _bankSelMethod(Normal)
    { }

    BankSelMethod bankSelMethod() const { return _bankSelMethod; }
    const QString& controls() const { return _controls; }
    const QString& nrpns() const { return _nrpns; }
    const QString& rpns() const { return _rpns; }
    const QMap<int, QString>& banks() const { return _banks; }
    const QMap<int, QMap<int, QString> >& keys() const { return _keys; }
    const QMap<int, QMap<int, bool> >& isDrums() const { return _isDrums; }
    const QValueVector<int>& patchList() const { return _patchList; }

    void setBankSelMethod(BankSelMethod method) { _bankSelMethod = method; }
    void setControls(const QString& controls) { _controls = controls; }
    void setNRPNs(const QString& nrpns) { _nrpns = nrpns;  }
    void setRPNs(const QString& rpns) { _rpns = rpns; }
    void setBank(int index, const QString& bank) { _banks[index] = bank; }
    void setKeys(int bank, int patch, const QString& keys) { _keys[bank][patch] = keys; }
    void setDrums(int bank, int patch, bool isDrums) { _isDrums[bank][patch] = isDrums; }

    void buildPatchList(const BankMap& banks);

  protected:
 
    BankSelMethod                  _bankSelMethod;
    QString                        _controls;
    QString                        _nrpns;
    QString                        _rpns;
    QMap<int, QString>             _banks;
    QMap<int, QMap<int, QString> > _keys;
    QMap<int, QMap<int, bool> >    _isDrums;

    QValueVector<int>              _patchList;
  };

  typedef QMap<QString, Instrument> InstrumentMap;

  class PortInfo
  {
  public:

    enum Type
    {
      Raw = 0, ALSA = 1, aRts = 2
    };

    Type type;

    QString name;
    bool    input;
    bool    output;
    int     client, port;
  };

  typedef QValueList<PortInfo> PortInfos;

  class Port
  {
  public:

    class Device
    {
    public:

      const QString name() const { return _name; }
      const QString type() const { return _type; }

      void setName(const QString& name) { _name = name; }
      void setType(const QString& type) { _type = type; }

    private:

      QString _name;
      QString _type;
    };

    Port(PortInfo portInfo)
      : _info(portInfo)
    { }

    PortInfo::Type type() const { return _info.type; }
    const QString deviceFile() const { return _info.name; }
    const Device& device(int index) const { return _devices[index]; }
    int client() const { return _info.client; }
    int port() const { return _info.port; }
    bool input() const { return _info.input; }
    bool output() const { return _info.output; }
    PortInfo info() const { return _info; }

    Device& device(int index) { return _devices[index]; }

  protected:

    PortInfo      _info;
    Device        _devices[MIDI_CHANNELS_MAX];
  };

  uint              portCount() const { return _ports.size(); }
  const Port&       port(uint i) const { return *_ports[i]; }
  const Bank&       bank(const QString& key) const { return _banks[key]; }
  const Control&    controls(const QString& key) const { return _controls[key]; }
  const NRPNs&      nrpns(const QString& key) const { return _nrpns[key]; }
  const RPNs&       rpns(const QString& key) const { return _rpns[key]; }
  const Notes&      notes(const QString& key) const { return _notes[key]; }
  const Instrument& instrument(const QString& key) const { return _instruments[key]; }

  /**
   * Perform automatic configuration (create basic MIDI device descriptions
   * and find MIDI ports).
   */
  void autoconfig();

protected:

  KConfig*         _kconfig;
  SanetrackerApp*  _parent;
  ConfigDialog*    _dialog;

  BankMap          _banks;
  ControlMap       _controls;
  NRPNMap          _nrpns;
  RPNMap           _rpns;
  NoteMap          _notes;
  InstrumentMap    _instruments;

  QPtrVector<Port> _ports;
  PortInfos        _portInfos;

  PortInfos        _portNameValues;
  PortInfos        _listPortsValues;

  /**
   * Read an instrument definition file.
   */
  void readInstrDef(const KURL& url, bool showErrors = true);

  /**
   * Write an instrument definition file.
   */
  void writeInstrDef(const KURL& url);

protected slots:

  void slotOkClicked();
  void slotApplyClicked();
  void slotCancelClicked();
  void slotDestroyed();
  void slotTypeURLChanged(const QString& url);
  void slotImportTypeClicked();
  void slotPortTypeActivated(int num);
  void slotPortListSelChanged();
  void slotAddPortClicked();
  void slotRemovePortClicked();
  void slotDeviceListSelChanged();
  void slotAddDeviceClicked();
  void slotRemoveDeviceClicked();
};

#endif

