#include <kapp.h>
#include <klocale.h>
#include <kaboutdata.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kmainwindow.h>
#include <kcmdlineargs.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qlayout.h>
#include <qvbox.h>
#include <qscrollbar.h>
#include <qtimer.h>
#include <stdio.h>
#include "artstracker.h"
#include "trackerview.h"
#include "main.h"

using namespace Arts::Tracker;
using namespace std;

ArtsTrackerWidget::ArtsTrackerWidget(QWidget *parent, QScrollBar *hbar,
	QScrollBar *vbar) : QWidget(parent), font( "courier", 18, QFont::Bold ),
	hbar(hbar), vbar(vbar)
{
	MidiManager manager = Reference("global:Arts_MidiManager");
	if(!manager.isNull())
	{
		midiClient = manager.addClient(Arts::mcdPlay,Arts::mctApplication,
									i18n("aRts Tracker").ascii(),"artstracker");
		midiPort = midiClient.addOutputPort();
	}
	else
		midiPort = MidiPort::null();

	hoffset = voffset = 0;
	notes = 64;
	cursorPattern = 0;
	cursorTrack = 0;
	cursorNote = 0;
	cursorOffset = 0;
	cursorSkip = 1;
	octave = 5;
	instrument = 0;

	p = song.getPattern(cursorPattern);
	tracks = p.trackCount();

	setBackgroundMode(NoBackground);
	setFocusPolicy(StrongFocus);
	calcSizes();
}

void ArtsTrackerWidget::calcSizes()
{
	QPainter painter;
	painter.begin(this);
	painter.setFont( font );
	trackWidth = painter.fontMetrics().width("C-4 06 .. ...  ");
	noteHeight = painter.fontMetrics().height();
	painter.end();
}

void ArtsTrackerWidget::changeHBar(int newValue)
{
	hoffset = newValue;
	paintNotes(0,0,notes,tracks,true);
}

void ArtsTrackerWidget::changeVBar(int newValue)
{
	int delta = voffset - newValue;
	voffset = newValue;
	scroll(0,delta * noteHeight);
}

void ArtsTrackerWidget::paintEvent(QPaintEvent *event)
{

  	QRect rect( event->rect() );
  	if ( rect.left() < 0 ) rect.rLeft() = 0;
  	if ( rect.right() > width() ) rect.rRight() = width();
  	if ( rect.top() < 0 ) rect.rTop() = 0;
  	if ( rect.bottom() > height() ) rect.rBottom() = height();

	paintNotes(0,0,notes,tracks,true,&rect);
}

void ArtsTrackerWidget::paintNotes(int startNote, int startTrack,
                                   int endNote, int endTrack, bool eraseBorder,
								   QRect *rect)
{
	QPainter painter;
	painter.begin(this);

	int minx = width(), maxx = 0, miny = height(), maxy = 0;
	int dx, dy;

	if(startNote < voffset) startNote = voffset;
	if(startTrack < hoffset) startTrack = hoffset;
	if(endNote > notes) endNote = notes;
	if(endTrack > tracks) endTrack = tracks;

	painter.setFont( font );
	painter.setPen( Qt::black );
	painter.setBackgroundMode( QPainter::OpaqueMode );
	painter.setBackgroundColor( Qt::white );

	int topMargin = 0; /*2 * noteHeight;*/
	int leftMargin = painter.fontMetrics().width("00 ");
	int charWidth = painter.fontMetrics().width("C");

	int t;
	for(t = startTrack; t < endTrack; t++)
	{
		Track track = p.getTrack(t);
		vector<Note> *notes = track.notes();
		assert(notes->size() == this->notes);

		int n;
		dx = (t-hoffset) * trackWidth + leftMargin;

		for(n = startNote;n < endNote; n++)
		{
			Note note = (*notes)[n];
			dy = (n-voffset) * noteHeight + topMargin;
			QRect noteRect(dx,dy,trackWidth,noteHeight);

			bool draw = true;

			if(rect)
				draw = rect->intersects(noteRect);

			/* check what is the min/max x/y we draw */
			if(dx < minx) minx = dx;
			if(dy < miny) miny = dy;
			if(dx+trackWidth > maxx) maxx = dx+trackWidth;
			if(dy+noteHeight > maxy) maxy = dy+noteHeight;

			if(draw)
			{
				if((n & 0xf) == 0)
					painter.setBackgroundColor( QColor(210,210,210) );
				else if((n & 0x3) == 0)
					painter.setBackgroundColor( QColor(235,235,235) );
				else
					painter.setBackgroundColor( white );
	
				painter.fillRect(noteRect, painter.backgroundColor());
				if(note.note == 255)
				{
					painter.drawText(noteRect,0,"[=] 00 .. ...  ");
				}
				else if(note.note != 0)
				{
					QString text;
					const char *nname[12] =
				  	{ "C-","C#","D-","D#","E-","F-","F#","G-","G#","A-","A#","H-" };
	
					text.sprintf("%s%x %02x %02x %x%02x  ",
						nname[note.note % 12], note.note/12, note.instrument,
						note.velocity, note.effect, note.effectParam);
	
					painter.drawText(noteRect,0,text);
				}
				else
				{
					painter.drawText(noteRect,0,"--- 00 .. ...  ");
				}
				if(n == cursorNote && t == cursorTrack)
				{
					int w = (cursorOffset == 0)?3*charWidth:1*charWidth;
	
					painter.setPen(red);
					painter.drawRect(dx+charWidth*cursorOffset,dy,w,noteHeight);
					painter.setPen(black);
				}
			}

			// paint margin
			if(t == hoffset)
			{
				QString number;
				number.sprintf("%02x ",n);
				painter.fillRect(0,dy,3*charWidth,noteHeight,painter.backgroundColor());
				painter.drawText(0,dy,3*charWidth,noteHeight,0,number);
			}
		}
		delete notes;

		painter.setPen(blue);
		painter.drawLine(dx+trackWidth-charWidth,miny,dx+trackWidth-charWidth,maxy);
		painter.setPen(black);
	}
	if(eraseBorder)
	{
		painter.fillRect(0,0,width(),miny,white);
		painter.fillRect(0,maxy,width(),maxy-miny,white);
		painter.fillRect(maxx,0,maxx-minx,height(),white);
	}
	painter.end();
}

int ArtsTrackerWidget::asPitch(QKeyEvent *event)
{
                  //     c   e  g    c   e  g
	const char *keys  = "q2w3er5t6z7ui9o0p'+";
	const char *lkeys = "ysxdcvgbhnjm,l.-";
	int found = -1,i;
	for(i=0;keys[i];i++)
		if(event->ascii() == keys[i]) found = octave*12+i;
	for(i=0;lkeys[i];i++)
		if(event->ascii() == lkeys[i]) found = (octave-1)*12+i;

	return found;
}

void ArtsTrackerWidget::keyReleaseEvent( QKeyEvent *event )
{
	int pitch = asPitch(event);

	if(pitch != -1 && !midiPort.isNull())
		midiPort.processCommand(MidiCommand(mcsNoteOff,pitch,0));
}

void ArtsTrackerWidget::keyPressEvent ( QKeyEvent* event )
{
	Track track = p.getTrack(cursorTrack);
	Note note = track.getNote(cursorNote);

	bool write = false;

	switch ( event->key() )
	{
		case Key_Plus:
				if(event->state() & AltButton)
				{
					cursorSkip++;
					event->accept();
				}
			break;
		case Key_Minus:
				if(event->state() & AltButton)
				{
					cursorSkip--;
					if(cursorSkip < 0) cursorSkip = 0;
					event->accept();
				}
			break;
    	case Key_Up:
				cursorMove(0,-1);
				event->accept();
			break;
    	case Key_Down:
				cursorMove(0,1);
				event->accept();
			break;
    	case Key_Left:
				cursorMove(-1,0);
				event->accept();
			break;
    	case Key_Right:
				cursorMove(1,0);
				event->accept();
			break;
    	case Key_PageUp:
				cursorMove(0,-4);
				event->accept();
			break;
    	case Key_PageDown:
				cursorMove(0,4);
				event->accept();
			break;
		case Key_F1:	// won't work, as this is also the hotkey for help
				setOctave(1);
				event->accept();
			break;
		case Key_F2:
				setOctave(2);
				event->accept();
			break;
		case Key_F3:
				setOctave(3);
				event->accept();
			break;
		case Key_F4:
				setOctave(4);
				event->accept();
			break;
		case Key_F5:
				setOctave(5);
				event->accept();
			break;
		case Key_F6:
				setOctave(6);
				event->accept();
			break;
		case Key_F7:
				setOctave(7);
				event->accept();
			break;
		case Key_F8:
				setOctave(8);
				event->accept();
			break;
		case Key_F9:
				setOctave(9);
				event->accept();
			break;
    	case Key_Delete:
				if(cursorOffset == 0)
				{
					note.note = 0;
					write = true;
				}
				event->accept();
			break;
	}

	if(cursorOffset == 0 && event->state() == 0)
	{
		int pitch = asPitch(event);
		
		if(pitch >= 0)
		{
			if(!midiPort.isNull())
			{
				midiPort.processCommand(MidiCommand(mcsProgram,instrument,0));
				midiPort.processCommand(MidiCommand(mcsNoteOn,pitch,100));
			}
			note.instrument = instrument;
			note.note = pitch;
			write = true;
		}
	}

	switch(event->ascii())
	{
		case 'P':
				emit togglePlay();
				event->accept();
			break;
		case '<':
				if(cursorOffset == 0)
				{
					note.note = 255;
					write = true;
				}
			break;
		case 'D':
				setPattern(cursorPattern+1);
				event->accept();
			break;
		case 'U':
				if(cursorPattern > 0) setPattern(cursorPattern-1);
				event->accept();
			break;
	}

	if(write)
	{
		track.setNote(cursorNote,note);
		paintNotes(cursorNote,cursorTrack,cursorNote+1,cursorTrack+1);
		cursorMove(0,cursorSkip);
	}
}

void ArtsTrackerWidget::cursorMove(int deltax, int deltay)
{
	int newCursorTrack = cursorTrack, newCursorNote = cursorNote,
	    newCursorOffset = cursorOffset;

	const char *layout = ".-4 .. .. ...";		// dots are places where the cursor may be
	do {
		newCursorOffset += deltax;
		if(newCursorOffset < 0)
		{
			newCursorOffset = strlen(layout);
			newCursorTrack--;
		}
		else if(newCursorOffset == strlen(layout))
		{
			newCursorOffset = 0;
			newCursorTrack++;
		}
	} while(layout[newCursorOffset] != '.');
	newCursorNote += deltay;

	// wrapping
	while(newCursorNote < 0) newCursorNote += notes;
	while(newCursorNote >= notes) newCursorNote -= notes;

	if(newCursorTrack < 0 || newCursorTrack >= tracks) return;

	int startNote = min(cursorNote, newCursorNote);
	int endNote = max(cursorNote, newCursorNote)+1;
	int startTrack = min(cursorTrack, newCursorTrack);
	int endTrack = max(cursorTrack, newCursorTrack)+1;

	cursorNote = newCursorNote;
	cursorTrack = newCursorTrack;
	cursorOffset = newCursorOffset;

	int screenNotes = height()/noteHeight - 2;

	int vnew = cursorNote - screenNotes / 2;
	if(vnew > notes - screenNotes) vnew = notes - screenNotes;
	if(vnew < 0) vnew = 0;

	vbar->setValue(vnew);

	while((cursorTrack-hoffset+1) * trackWidth > width())
		hbar->setValue(hbar->value()+1);
	while(cursorTrack < hoffset)
		hbar->setValue(hbar->value()-1);
	paintNotes(startNote,startTrack,endNote, endTrack);
}

void ArtsTrackerWidget::setCursorNote(int newNote)
{
	if(cursorNote == newNote) return;

	cursorMove(0,newNote-cursorNote);
}

void ArtsTrackerWidget::setPattern(int newPattern)
{
	if(cursorPattern == newPattern) return;
	cursorPattern = newPattern;

	p = song.getPattern(cursorPattern);
	redrawNotes();

	emit patternChanged(newPattern);
}

void ArtsTrackerWidget::redrawNotes()
{
	paintNotes(0,0,notes,tracks,true);
}

void ArtsTrackerWidget::songChanged()
{
	/* force redraw */
	cursorPattern = 1;
	cursorNote = 0;
	cursorOffset = 0;
	setPattern(0);
}

void ArtsTrackerWidget::setInstrument(int newInstrument)
{
	if(instrument != newInstrument)
	{
		instrument = newInstrument;
		emit instrumentChanged(newInstrument);
	}
}

void ArtsTrackerWidget::setOctave(int newOctave)
{
	if(octave != newOctave)
	{
		octave = newOctave;
		emit octaveChanged(newOctave);
	}
}

SongPlayer::SongPlayer(QObject *parent, Song song, MidiPort port)
	: QObject(parent), song(song), midiPort(port)
{
	playTimer = new QTimer(this);
	playing = false;
	connect( playTimer, SIGNAL(timeout()), this, SLOT(timerTick()) );
}

SongPlayer::~SongPlayer()
{
	if(playing) togglePlay();
}

void SongPlayer::togglePlay()
{
	currentSongPosition = 0;
	Pattern p = song.getPattern(song.getSongPattern(currentSongPosition));

	if(playing)
	{
		for(int t = 0; t < p.trackCount(); t++)
		{
			if(activeNote[t] != -1)
			{
				midiPort.processEvent(
					MidiEvent(playTime,
						MidiCommand(mcsNoteOff|t,activeNote[t],0)));
			}
		}
		playTimer->stop();

		while(!playQueue.empty()) playQueue.pop();
	}
	else
	{
		playTime = midiPort.time();
		playTime.usec += 500000;   // 500ms initialization prerun
		playTime.sec += playTime.usec / 1000000;
		playTime.usec %= 1000000;
		currentPosition = 0;
		playTimer->start(60,false);

		activeNote.resize(p.trackCount());
		activeInstrument.resize(p.trackCount());
		for(int t = 0; t < p.trackCount(); t++)
			activeNote[t] = activeInstrument[t] = -1;
	}
	playing = !playing;
}

float deltaT(TimeStamp a, TimeStamp b)
{
	return ((float)(a.sec - b.sec) + (float)(a.usec - b.usec)/1000000.0);
}

void SongPlayer::timerTick()
{
	TimeStamp midiTime = midiPort.time();

	while(deltaT(playTime,midiTime) < 1.000) //0.300)
	{
		long currentPattern = song.getSongPattern(currentSongPosition); 
		Pattern p = song.getPattern(currentPattern);

		playQueue.push(currentPattern);
		playQueue.push(currentPosition);

		int t;
		for(t = 0; t < p.trackCount(); t++)
		{
			Track track = p.getTrack(t);
			Note note = track.getNote(currentPosition);
			if(note.note != 0)
			{
				MidiEvent ev;
				ev.time = playTime;

				if(activeNote[t] != -1)
				{
					ev.command = MidiCommand(mcsNoteOff|t,activeNote[t],0);
					midiPort.processEvent(ev);
				}
	
				if(note.note == 255)
				{
					activeNote[t] = -1;
				}
				else
				{
					// do instrument switch when necessary
					if(activeInstrument[t] != note.instrument)
					{
						activeInstrument[t] = note.instrument;
						ev.command
							= MidiCommand(mcsProgram|t,activeInstrument[t],0);
						midiPort.processEvent(ev);
					}

					activeNote[t] = note.note;
					ev.command = MidiCommand(mcsNoteOn|t,activeNote[t],100);
					midiPort.processEvent(ev);
				}
			}
		}
		currentPosition++;
		if(currentPosition >= 64)
		{
			currentPosition = 0;
			currentSongPosition++;
			if(currentSongPosition >= song.songLength())
				currentSongPosition = 0;
		}
		playTime.usec += 60000;   // 60ms
		playTime.sec += playTime.usec / 1000000;
		playTime.usec %= 1000000;
	}
	while(playQueue.size() > 12)
	{
		playQueue.pop();
		playQueue.pop();
	}
	if(playQueue.size() > 10)
	{
		emit playPattern(playQueue.front());
		playQueue.pop();

		emit playNote(playQueue.front());
		playQueue.pop();
	}
}

void SongPlayer::stopAllNotes()
{
	if(!midiPort.isNull())
	{
		for(int t=0;t<16;t++)
		{
			// all notes off midi event
			midiPort.processCommand(MidiCommand(0xb0|t,0x7b,0));
		}
	}
}

bool SongPlayer::isPlaying()
{
	return playing;
}
	
class MainWindow : public KMainWindow {
protected:
	TrackerView *main;
public:
	MainWindow()
	{
		main = new TrackerView(this);
		setCentralWidget(main);
		KStdAction::openNew(main, SLOT(file_new()), actionCollection());
		KStdAction::open(main, SLOT(file_open()), actionCollection());
    	KStdAction::save(main, SLOT(file_save()), actionCollection());
    	KStdAction::quit(this, SLOT(close()), actionCollection());
		(void)new KAction(i18n("&Play Song"), ALT+Key_P, main,
				SLOT(file_play_song()), actionCollection(), "file_play_song");
		(void)new KAction(i18n("Stop all notes"), ALT+Key_Space, main,
				SLOT(file_stop_all_notes()), actionCollection(), "file_stop_all_notes");
		createGUI("artstracker.rc");
		resize(600,400);
	}
};

static const char *description=I18N_NOOP("aRts tracker");

int main(int argc, char **argv)
{
	KAboutData aboutData( "artstracker", I18N_NOOP("aRts tracker"),
	   "0.1", description, KAboutData::License_GPL, "(c) 2000 Stefan Westerfeld");
 
	KCmdLineArgs::init( argc, argv, &aboutData );

	KApplication app;
	Arts::Dispatcher dispatcher;
	MainWindow *w = new MainWindow;
	w->show();
	app.setMainWidget(w);
	return app.exec();
}
